# Copyright 2017 Pilosa Corp.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived
# from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
# DAMAGE.
#

from .exceptions import PilosaError
from .internal import public_pb2 as internal


__all__ = ("BitmapResult", "CountResultItem", "QueryResult", "ColumnItem", "QueryResponse")


class BitmapResult:
    """Represents a result from ``Bitmap``, ``Union``, ``Intersect``, ``Difference`` and ``Range`` PQL calls.
    
    * See `Query Language <https://www.pilosa.com/docs/query-language/>`_
    """

    def __init__(self, bits=None, attributes=None):
        self.bits = bits or []
        self.attributes = attributes or {}

    @classmethod
    def from_internal(cls, obj):
        return cls(list(obj.Bits), _convert_protobuf_attrs_to_dict(obj.Attrs))


class CountResultItem:
    """Represents a result from ``TopN`` call.

    * See `Query Language <https://www.pilosa.com/docs/query-language/>`_    
    """

    def __init__(self, id, count):
        self.id = id
        self.count = count


class QueryResult:
    """Represents one of the results in the response.
    
    * See `Query Language <https://www.pilosa.com/docs/query-language/>`_        
    """

    def __init__(self, bitmap=None, count_items=None, count=0, sum=0):
        self.bitmap = bitmap or BitmapResult()
        self.count_items = count_items or []
        self.count = count
        self.sum = sum

    @classmethod
    def from_internal(cls, obj):
        count_items = []
        for pair in obj.Pairs:
            count_items.append(CountResultItem(pair.Key, pair.Count))
        count = obj.N if obj.N > 0 else obj.SumCount.Count
        return cls(BitmapResult.from_internal(obj.Bitmap),
                   count_items,
                   count,
                   obj.SumCount.Sum)


class ColumnItem:
    """Contains data about a column.
    
    Column data is returned from ``QueryResponse.getColumns()`` method.
    They are only returned if ``Client.query`` was called with ``columns=True``.
 """

    def __init__(self, id, attributes):
        self.id = id
        self.attributes = attributes

    @classmethod
    def _from_internal(cls, obj):
        return cls(obj.ID, _convert_protobuf_attrs_to_dict(obj.Attrs))


class QueryResponse(object):
    """Represents the response from a Pilosa query.

    * See `Query Language <https://www.pilosa.com/docs/query-language/>`_        
    """

    def __init__(self, results=None, columns=None, error_message=""):
        self.results = results or []
        self.columns = columns or []
        self.error_message = error_message

    @classmethod
    def _from_protobuf(cls, bin):
        response = internal.QueryResponse()
        response.ParseFromString(bin)
        results = [QueryResult.from_internal(r) for r in response.Results]
        columns = [ColumnItem._from_internal(p) for p in response.ColumnAttrSets]
        error_message = response.Err
        return cls(results, columns, error_message)

    @property
    def result(self):
        return self.results[0] if self.results else None

    @property
    def column(self):
        return self.columns[0] if self.columns else None


def _convert_protobuf_attrs_to_dict(attrs):
    protobuf_attrs_to_dict = [
        None,
        lambda a: a.StringValue,
        lambda a: a.IntValue,
        lambda a: a.BoolValue,
        lambda a: a.FloatValue,
    ]
    d = {}
    attr = None  # to get the attr with invalid type
    try:
        for attr in attrs:
            value = protobuf_attrs_to_dict[attr.Type](attr)
            d[attr.Key] = value
    except (IndexError, TypeError):
        raise PilosaError("Invalid protobuf attribute type: %s" % attr.Type)
    return d



# Copyright 2017 Pilosa Corp.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived
# from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
# DAMAGE.
#

import json
import logging
import re
import threading

import urllib3

from .exceptions import PilosaError, PilosaURIError, IndexExistsError, FrameExistsError
from .imports import batch_bits
from .internal import public_pb2 as internal
from .orm import TimeQuantum, Schema, CacheType
from .response import QueryResponse
from .version import VERSION

__all__ = ("Client", "Cluster", "URI")

_LOGGER = logging.getLogger("pilosa")
_MAX_HOSTS = 10


class Client(object):
    """Pilosa HTTP client

    This client uses Pilosa's http+protobuf API.

    Usage: ::

        import pilosa

        # Create a Client instance
        client = pilosa.Client()

        # Create an Index instance
        index = pilosa.Index("repository")

        stargazer = index.frame("stargazer")
        response = client.query(stargazer.bitmap(5))

        # Act on the result
        print(response.result)

    * See `Pilosa API Reference <https://www.pilosa.com/docs/api-reference/>`_.
    * See `Query Language <https://www.pilosa.com/docs/query-language/>`_.
    """

    __NO_RESPONSE, __RAW_RESPONSE, __ERROR_CHECKED_RESPONSE = range(3)

    def __init__(self, cluster_or_uri=None, connect_timeout=30000, socket_timeout=300000,
                 pool_size_per_route=10, pool_size_total=100, retry_count=3,
                 tls_skip_verify=False, tls_ca_certificate_path=""):
        if cluster_or_uri is None:
            self.cluster = Cluster(URI())
        elif isinstance(cluster_or_uri, Cluster):
            self.cluster = cluster_or_uri.copy()
        elif isinstance(cluster_or_uri, URI):
            self.cluster = Cluster(cluster_or_uri)
        elif isinstance(cluster_or_uri, str):
            self.cluster = Cluster(URI.address(cluster_or_uri))
        else:
            raise PilosaError("Invalid cluster_or_uri: %s" % cluster_or_uri)

        self.connect_timeout = connect_timeout / 1000.0
        self.socket_timeout = socket_timeout / 1000.0
        self.pool_size_per_route = pool_size_per_route
        self.pool_size_total = pool_size_total
        self.retry_count = retry_count
        self.tls_skip_verify = tls_skip_verify
        self.tls_ca_certificate_path = tls_ca_certificate_path
        self.__current_host = None
        self.__client = None

    def query(self, query, columns=False, exclude_bits=False, exclude_attrs=False):
        """Runs the given query against the server with the given options.
        
        :param pilosa.PqlQuery query: a PqlQuery object with a non-null index
        :param bool columns: Enables returning column data from bitmap queries
        :param bool exclude_bits: Disables returning bits from bitmap queries
        :param bool exclude_attrs: Disables returning attributes from bitmap queries
        :return: Pilosa response
        :rtype: pilosa.Response
        """
        request = _QueryRequest(query.serialize(), columns=columns, exclude_bits=exclude_bits, exclude_attrs=exclude_attrs)
        data = bytearray(request.to_protobuf())
        path = "/index/%s/query" % query.index.name
        response = self.__http_request("POST", path, data=data, client_response=Client.__RAW_RESPONSE)
        query_response = QueryResponse._from_protobuf(response.data)
        if query_response.error_message:
            raise PilosaError(query_response.error_message)
        return query_response

    def create_index(self, index):
        """Creates an index on the server using the given Index object.
        
        :param pilosa.Index index:
        :raises pilosa.IndexExistsError: if there already is a index with the given name
        """
        data = json.dumps({
            "options": {"columnLabel": index.column_label}
        })
        path = "/index/%s" % index.name
        self.__http_request("POST", path, data=data)
        if index.time_quantum != TimeQuantum.NONE:
            self.__patch_index_time_quantum(index)

    def delete_index(self, index):
        """Deletes the given index on the server.
        
        :param pilosa.Index index:
        :raises pilosa.PilosaError: if the index does not exist
        """
        path = "/index/%s" % index.name
        self.__http_request("DELETE", path)

    def create_frame(self, frame):
        """Creates a frame on the server using the given Frame object.
        
        :param pilosa.Frame frame:
        :raises pilosa.FrameExistsError: if there already is a frame with the given name
        """
        data = frame._get_options_string()
        path = "/index/%s/frame/%s" % (frame.index.name, frame.name)
        self.__http_request("POST", path, data=data)

    def delete_frame(self, frame):
        """Deletes the given frame on the server.
        
        :param pilosa.Frame frame:
        :raises pilosa.PilosaError: if the frame does not exist
        """
        path = "/index/%s/frame/%s" % (frame.index.name, frame.name)
        self.__http_request("DELETE", path)

    def ensure_index(self, index):
        """Creates an index on the server if it does not exist.
        
        :param pilosa.Index index:
        """
        try:
            self.create_index(index)
        except IndexExistsError:
            pass

    def ensure_frame(self, frame):
        """Creates a frame on the server if it does not exist.
        
        :param pilosa.Frame frame:
        """
        try:
            self.create_frame(frame)
        except FrameExistsError:
            pass

    def status(self):
        response = self.__http_request("GET", "/status",
                                       client_response=Client.__ERROR_CHECKED_RESPONSE)
        return json.loads(response.data.decode('utf-8'))["status"]

    def schema(self):
        status = self.status()
        nodes = status.get("Nodes")
        schema = Schema()
        for index_info in nodes[0].get("Indexes", []):
            options = decode_index_meta_options(index_info)
            index = schema.index(index_info["Name"], **options)
            for frame_info in index_info.get("Frames", []):
                options = decode_frame_meta_options(frame_info)
                index.frame(frame_info["Name"], **options)

        return schema

    def sync_schema(self, schema):
        server_schema = self.schema()

        # find out local - remote schema
        diff_schema = schema._diff(server_schema)
        # create indexes and frames which doesn't exist on the server side
        for index_name, index in diff_schema._indexes.items():
            if index_name not in server_schema._indexes:
                self.ensure_index(index)
            for frame_name, frame in index._frames.items():
                self.ensure_frame(frame)

        # find out remote - local schema
        diff_schema = server_schema._diff(schema)
        for index_name, index in diff_schema._indexes.items():
            local_index = schema._indexes.get(index_name)
            if local_index is None:
                schema._indexes[index_name] = index
            else:
                for frame_name, frame in index._frames.items():
                    local_index._frames[frame_name] = frame

    def import_frame(self, frame, bit_reader, batch_size=100000):
        """Imports a frame using the given bit reader

        :param frame:
        :param bit_reader:
        :param batch_size:
        """
        index_name = frame.index.name
        frame_name = frame.name
        import_bits = self._import_bits
        for slice, bits in batch_bits(bit_reader, batch_size):
            import_bits(index_name, frame_name, slice, bits)

    def http_request(self, method, path, data=None, headers=None):
        """Sends an HTTP request to the Pilosa server

        NOTE: This function is experimental and may be removed in later revisions.

        :param method: HTTP method
        :param path: Request path
        :param data: Request body
        :param headers: Request headers
        :return HTTP response:

        """
        return self.__http_request(method, path, data=data, headers=headers,
                                   client_response=Client.__RAW_RESPONSE)

    def _import_bits(self, index_name, frame_name, slice, bits):
        # sort by row_id then by column_id
        bits.sort(key=lambda bit: (bit.row_id, bit.column_id))
        nodes = self._fetch_fragment_nodes(index_name, slice)
        for node in nodes:
            client_params={
                "tls_skip_verify": self.tls_skip_verify,
                "tls_ca_certificate_path": self.tls_ca_certificate_path,
            }
            address = "%s://%s" % (node["scheme"], node["host"])
            client = Client(URI.address(address), **client_params)
            client._import_node(_ImportRequest(index_name, frame_name, slice, bits))

    def _fetch_fragment_nodes(self, index_name, slice):
        path = "/fragment/nodes?slice=%d&index=%s" % (slice, index_name)
        response = self.__http_request("GET", path, client_response=Client.__ERROR_CHECKED_RESPONSE)
        content = response.data.decode('utf-8')
        return json.loads(content)

    def _import_node(self, import_request):
        data = import_request.to_protobuf()
        self.__http_request("POST", "/import", data=data, client_response=Client.__RAW_RESPONSE)

    def __patch_index_time_quantum(self, index):
        path = "/index/%s/time-quantum" % index.name
        data = '{\"timeQuantum\":\"%s\"}"' % str(index.time_quantum)
        self.__http_request("PATCH", path, data=data)

    def __http_request(self, method, path, data=None, headers=None, client_response=0):
        if not self.__client:
            self.__connect()
        # try at most 10 non-failed hosts; protect against broken cluster.remove_host
        for _ in range(_MAX_HOSTS):
            uri = "%s%s" % (self.__get_address(), path)
            try:
                _LOGGER.debug("Request: %s %s %s", method, uri,
                              "[binary]" if client_response == Client.__RAW_RESPONSE else data)
                response = self.__client.request(method, uri, body=data, headers=headers)
                break
            except urllib3.exceptions.MaxRetryError as e:
                self.cluster.remove_host(self.__current_host)
                _LOGGER.warning("Removed %s from the cluster due to %s", self.__current_host, str(e))
                self.__current_host = None
        else:
            raise PilosaError("Tried %s hosts, still failing" % _MAX_HOSTS)

        if client_response == Client.__RAW_RESPONSE:
            return response

        if 200 <= response.status < 300:
            return None if client_response == Client.__NO_RESPONSE else response
        content = response.data.decode('utf-8')
        ex = self.__RECOGNIZED_ERRORS.get(content)
        if ex is not None:
            raise ex
        raise PilosaError("Server error (%d): %s" % (response.status, content))

    def __get_address(self):
        if self.__current_host is None:
            self.__current_host = self.cluster.get_host()
            _LOGGER.debug("Current host set: %s", self.__current_host)
        return self.__current_host._normalize()

    def __connect(self):
        num_pools = float(self.pool_size_total) / self.pool_size_per_route
        headers = {
            'Content-Type': 'application/x-protobuf',
            'Accept': 'application/x-protobuf',
            'User-Agent': 'python-pilosa/' + VERSION,
        }

        timeout = urllib3.Timeout(connect=self.connect_timeout, read=self.socket_timeout)
        client_options = {
            "num_pools": num_pools,
            "maxsize": self.pool_size_per_route,
            "block": True,
            "headers": headers,
            "timeout": timeout,
            "retries": self.retry_count,
        }
        if not self.tls_skip_verify:
            client_options["cert_reqs"] = "CERT_REQUIRED"
            client_options["ca_certs"] = self.tls_ca_certificate_path

        client = urllib3.PoolManager(**client_options)
        self.__client = client

    __RECOGNIZED_ERRORS = {
        "index already exists\n": IndexExistsError,
        "frame already exists\n": FrameExistsError,
    }


def decode_index_meta_options(index_info):
    meta = index_info.get("Meta", {})
    return {
        "column_label": meta.get("ColumnLabel", "columnID"),
        "time_quantum": TimeQuantum(meta.get("TimeQuantum", "")),
    }


def decode_frame_meta_options(frame_info):
    meta = frame_info.get("Meta", {})
    return {
        "row_label": meta.get("RowLabel", "rowID"),
        "cache_size": meta.get("CacheSize", 50000),
        "cache_type": CacheType(meta.get("CacheType", "")),
        "inverse_enabled": meta.get("InverseEnabled", False),
        "time_quantum": TimeQuantum(meta.get("TimeQuantum", "")),
    }


class URI:
    """Represents a Pilosa URI

    A Pilosa URI consists of three parts:

    * Scheme: Protocol of the URI. Default: ``http``
    * Host: Hostname or IP URI. Default: ``localhost``
    * Port: Port of the URI. Default ``10101``

    All parts of the URI are optional. The following are equivalent:

    * ``http://localhost:10101``
    * ``http://localhost``
    * ``http://:10101``
    * ``localhost:10101``
    * ``localhost``
    * ``:10101``

    :param str scheme: is the scheme of the Pilosa Server, such as ``http`` or ``https``
    :param str host: is the hostname or IP address of the Pilosa server. IPv6 addresses should be enclosed in brackets, e.g., ``[fe00::0]``.
    :param int port: is the port of the Pilosa server
    """
    __PATTERN = re.compile("^(([+a-z]+):\\/\\/)?([0-9a-z.-]+|\\[[:0-9a-fA-F]+\\])?(:([0-9]+))?$")

    def __init__(self, scheme="http", host="localhost", port=10101):
        self.scheme = scheme
        self.host = host
        self.port = port

    @classmethod
    def address(cls, address):
        """ Creates a URI from an address.
        
        :param str address: of the form ``${SCHEME}://${HOST}:{$PORT}``
        :return: a Pilosa URI
        :type: pilosa.URI
        """
        uri = cls()
        uri._parse(address)
        return uri

    def _normalize(self):
        scheme = self.scheme
        try:
            index = scheme.index("+")
            scheme = scheme[:index]
        except ValueError:
            pass
        return "%s://%s:%s" % (scheme, self.host, self.port)

    def _parse(self, address):
        m = self.__PATTERN.search(address)
        if m:
            scheme = m.group(2)
            if scheme:
                self.scheme = scheme
            host = m.group(3)
            if host:
                self.host = host
            port = m.group(5)
            if port:
                self.port = int(port)
            return
        raise PilosaURIError("Not a Pilosa URI")

    def __str__(self):
        return "%s://%s:%s" % (self.scheme, self.host, self.port)

    def __repr__(self):
        return "<URI %s>" % self

    def __eq__(self, other):
        if id(self) == id(other):
            return True
        if other is None or not isinstance(other, self.__class__):
            return False
        return self.scheme == other.scheme and \
            self.host == other.host and \
            self.port == other.port


class Cluster:
    """Contains hosts in a Pilosa cluster.
    
    :param hosts: URIs of hosts. Leaving out hosts creates the default cluster
    """

    def __init__(self, *hosts):
        """Returns the cluster with the given hosts"""
        self.hosts = [(host, True) for host in hosts]
        self.__next_index = 0
        self.__lock = threading.RLock()

    def add_host(self, uri):
        """Makes a host available.
        
        :param pilosa.URI uri:
        """
        with self.__lock:
            for i, item in enumerate(self.hosts):
                host, _ = item
                if host == uri:
                    self.hosts[i] = (host, True)
                    break
            else:
                self.hosts.append((uri, True))

    def remove_host(self, uri):
        """Makes a host unavailable.
        
        :param pilosa.URI uri:
        """
        with self.__lock:
            for i, item in enumerate(self.hosts):
                if uri == item[0]:
                    self.hosts[i] = (item[0], False)

    def get_host(self):
        """Returns the next host in the cluster.
        
        :return: next host
        :rtype: pilosa.URI         
        """
        for host, ok in self.hosts:
            if not ok:
                continue
            return host
        else:
            self._reset()
            raise PilosaError("There are no available hosts")

    def copy(self):
        c = Cluster()
        c.hosts = self.hosts[:]
        return c

    def _reset(self):
        with self.__lock:
            self.hosts = [(host, True) for host, _ in self.hosts]


class _QueryRequest:

    def __init__(self, query, columns=False, exclude_bits=False, exclude_attrs=False):
        self.query = query
        self.columns = columns
        self.exclude_bits = exclude_bits
        self.exclude_attrs = exclude_attrs

    def to_protobuf(self):
        qr = internal.QueryRequest()
        qr.Query = self.query
        qr.ColumnAttrs = self.columns
        qr.ExcludeBits = self.exclude_bits
        qr.ExcludeAttrs = self.exclude_attrs
        return qr.SerializeToString()


class _ImportRequest:

    def __init__(self, index_name, frame_name, slice, bits):
        self.index_name = index_name
        self.frame_name = frame_name
        self.slice = slice
        self.bits = bits

    def to_protobuf(self):
        import_request = internal.ImportRequest()
        import_request.Index = self.index_name
        import_request.Frame = self.frame_name
        import_request.Slice = self.slice
        row_ids = import_request.RowIDs
        column_ids = import_request.ColumnIDs
        timestamps = import_request.Timestamps
        for bit in self.bits:
            row_ids.append(bit.row_id)
            column_ids.append(bit.column_id)
            timestamps.append(bit.timestamp)
        return import_request.SerializeToString()




# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: public.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='public.proto',
  package='internal',
  syntax='proto3',
  serialized_pb=_b('\n\x0cpublic.proto\x12\x08internal\"5\n\x06\x42itmap\x12\x0c\n\x04\x42its\x18\x01 \x03(\x04\x12\x1d\n\x05\x41ttrs\x18\x02 \x03(\x0b\x32\x0e.internal.Attr\"\"\n\x04Pair\x12\x0b\n\x03Key\x18\x01 \x01(\x04\x12\r\n\x05\x43ount\x18\x02 \x01(\x04\"&\n\x08SumCount\x12\x0b\n\x03Sum\x18\x01 \x01(\x03\x12\r\n\x05\x43ount\x18\x02 \x01(\x03\"9\n\x03\x42it\x12\r\n\x05RowID\x18\x01 \x01(\x04\x12\x10\n\x08\x43olumnID\x18\x02 \x01(\x04\x12\x11\n\tTimestamp\x18\x03 \x01(\x03\":\n\rColumnAttrSet\x12\n\n\x02ID\x18\x01 \x01(\x04\x12\x1d\n\x05\x41ttrs\x18\x02 \x03(\x0b\x32\x0e.internal.Attr\"o\n\x04\x41ttr\x12\x0b\n\x03Key\x18\x01 \x01(\t\x12\x0c\n\x04Type\x18\x02 \x01(\x04\x12\x13\n\x0bStringValue\x18\x03 \x01(\t\x12\x10\n\x08IntValue\x18\x04 \x01(\x03\x12\x11\n\tBoolValue\x18\x05 \x01(\x08\x12\x12\n\nFloatValue\x18\x06 \x01(\x01\"(\n\x07\x41ttrMap\x12\x1d\n\x05\x41ttrs\x18\x01 \x03(\x0b\x32\x0e.internal.Attr\"}\n\x0cQueryRequest\x12\r\n\x05Query\x18\x01 \x01(\t\x12\x0e\n\x06Slices\x18\x02 \x03(\x04\x12\x13\n\x0b\x43olumnAttrs\x18\x03 \x01(\x08\x12\x0e\n\x06Remote\x18\x05 \x01(\x08\x12\x14\n\x0c\x45xcludeAttrs\x18\x06 \x01(\x08\x12\x13\n\x0b\x45xcludeBits\x18\x07 \x01(\x08\"u\n\rQueryResponse\x12\x0b\n\x03\x45rr\x18\x01 \x01(\t\x12&\n\x07Results\x18\x02 \x03(\x0b\x32\x15.internal.QueryResult\x12/\n\x0e\x43olumnAttrSets\x18\x03 \x03(\x0b\x32\x17.internal.ColumnAttrSet\"\x90\x01\n\x0bQueryResult\x12 \n\x06\x42itmap\x18\x01 \x01(\x0b\x32\x10.internal.Bitmap\x12\t\n\x01N\x18\x02 \x01(\x04\x12\x1d\n\x05Pairs\x18\x03 \x03(\x0b\x32\x0e.internal.Pair\x12$\n\x08SumCount\x18\x05 \x01(\x0b\x32\x12.internal.SumCount\x12\x0f\n\x07\x43hanged\x18\x04 \x01(\x08\"s\n\rImportRequest\x12\r\n\x05Index\x18\x01 \x01(\t\x12\r\n\x05\x46rame\x18\x02 \x01(\t\x12\r\n\x05Slice\x18\x03 \x01(\x04\x12\x0e\n\x06RowIDs\x18\x04 \x03(\x04\x12\x11\n\tColumnIDs\x18\x05 \x03(\x04\x12\x12\n\nTimestamps\x18\x06 \x03(\x03\x62\x06proto3')
)
_sym_db.RegisterFileDescriptor(DESCRIPTOR)




_BITMAP = _descriptor.Descriptor(
  name='Bitmap',
  full_name='internal.Bitmap',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='Bits', full_name='internal.Bitmap.Bits', index=0,
      number=1, type=4, cpp_type=4, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='Attrs', full_name='internal.Bitmap.Attrs', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=26,
  serialized_end=79,
)


_PAIR = _descriptor.Descriptor(
  name='Pair',
  full_name='internal.Pair',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='Key', full_name='internal.Pair.Key', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='Count', full_name='internal.Pair.Count', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=81,
  serialized_end=115,
)


_SUMCOUNT = _descriptor.Descriptor(
  name='SumCount',
  full_name='internal.SumCount',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='Sum', full_name='internal.SumCount.Sum', index=0,
      number=1, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='Count', full_name='internal.SumCount.Count', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=117,
  serialized_end=155,
)


_BIT = _descriptor.Descriptor(
  name='Bit',
  full_name='internal.Bit',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='RowID', full_name='internal.Bit.RowID', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='ColumnID', full_name='internal.Bit.ColumnID', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='Timestamp', full_name='internal.Bit.Timestamp', index=2,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=157,
  serialized_end=214,
)


_COLUMNATTRSET = _descriptor.Descriptor(
  name='ColumnAttrSet',
  full_name='internal.ColumnAttrSet',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='ID', full_name='internal.ColumnAttrSet.ID', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='Attrs', full_name='internal.ColumnAttrSet.Attrs', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=216,
  serialized_end=274,
)


_ATTR = _descriptor.Descriptor(
  name='Attr',
  full_name='internal.Attr',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='Key', full_name='internal.Attr.Key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='Type', full_name='internal.Attr.Type', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='StringValue', full_name='internal.Attr.StringValue', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='IntValue', full_name='internal.Attr.IntValue', index=3,
      number=4, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='BoolValue', full_name='internal.Attr.BoolValue', index=4,
      number=5, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='FloatValue', full_name='internal.Attr.FloatValue', index=5,
      number=6, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=276,
  serialized_end=387,
)


_ATTRMAP = _descriptor.Descriptor(
  name='AttrMap',
  full_name='internal.AttrMap',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='Attrs', full_name='internal.AttrMap.Attrs', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=389,
  serialized_end=429,
)


_QUERYREQUEST = _descriptor.Descriptor(
  name='QueryRequest',
  full_name='internal.QueryRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='Query', full_name='internal.QueryRequest.Query', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='Slices', full_name='internal.QueryRequest.Slices', index=1,
      number=2, type=4, cpp_type=4, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='ColumnAttrs', full_name='internal.QueryRequest.ColumnAttrs', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='Remote', full_name='internal.QueryRequest.Remote', index=3,
      number=5, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='ExcludeAttrs', full_name='internal.QueryRequest.ExcludeAttrs', index=4,
      number=6, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='ExcludeBits', full_name='internal.QueryRequest.ExcludeBits', index=5,
      number=7, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=431,
  serialized_end=556,
)


_QUERYRESPONSE = _descriptor.Descriptor(
  name='QueryResponse',
  full_name='internal.QueryResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='Err', full_name='internal.QueryResponse.Err', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='Results', full_name='internal.QueryResponse.Results', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='ColumnAttrSets', full_name='internal.QueryResponse.ColumnAttrSets', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=558,
  serialized_end=675,
)


_QUERYRESULT = _descriptor.Descriptor(
  name='QueryResult',
  full_name='internal.QueryResult',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='Bitmap', full_name='internal.QueryResult.Bitmap', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='N', full_name='internal.QueryResult.N', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='Pairs', full_name='internal.QueryResult.Pairs', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='SumCount', full_name='internal.QueryResult.SumCount', index=3,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='Changed', full_name='internal.QueryResult.Changed', index=4,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=678,
  serialized_end=822,
)


_IMPORTREQUEST = _descriptor.Descriptor(
  name='ImportRequest',
  full_name='internal.ImportRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='Index', full_name='internal.ImportRequest.Index', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='Frame', full_name='internal.ImportRequest.Frame', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='Slice', full_name='internal.ImportRequest.Slice', index=2,
      number=3, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='RowIDs', full_name='internal.ImportRequest.RowIDs', index=3,
      number=4, type=4, cpp_type=4, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='ColumnIDs', full_name='internal.ImportRequest.ColumnIDs', index=4,
      number=5, type=4, cpp_type=4, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='Timestamps', full_name='internal.ImportRequest.Timestamps', index=5,
      number=6, type=3, cpp_type=2, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=824,
  serialized_end=939,
)

_BITMAP.fields_by_name['Attrs'].message_type = _ATTR
_COLUMNATTRSET.fields_by_name['Attrs'].message_type = _ATTR
_ATTRMAP.fields_by_name['Attrs'].message_type = _ATTR
_QUERYRESPONSE.fields_by_name['Results'].message_type = _QUERYRESULT
_QUERYRESPONSE.fields_by_name['ColumnAttrSets'].message_type = _COLUMNATTRSET
_QUERYRESULT.fields_by_name['Bitmap'].message_type = _BITMAP
_QUERYRESULT.fields_by_name['Pairs'].message_type = _PAIR
_QUERYRESULT.fields_by_name['SumCount'].message_type = _SUMCOUNT
DESCRIPTOR.message_types_by_name['Bitmap'] = _BITMAP
DESCRIPTOR.message_types_by_name['Pair'] = _PAIR
DESCRIPTOR.message_types_by_name['SumCount'] = _SUMCOUNT
DESCRIPTOR.message_types_by_name['Bit'] = _BIT
DESCRIPTOR.message_types_by_name['ColumnAttrSet'] = _COLUMNATTRSET
DESCRIPTOR.message_types_by_name['Attr'] = _ATTR
DESCRIPTOR.message_types_by_name['AttrMap'] = _ATTRMAP
DESCRIPTOR.message_types_by_name['QueryRequest'] = _QUERYREQUEST
DESCRIPTOR.message_types_by_name['QueryResponse'] = _QUERYRESPONSE
DESCRIPTOR.message_types_by_name['QueryResult'] = _QUERYRESULT
DESCRIPTOR.message_types_by_name['ImportRequest'] = _IMPORTREQUEST

Bitmap = _reflection.GeneratedProtocolMessageType('Bitmap', (_message.Message,), dict(
  DESCRIPTOR = _BITMAP,
  __module__ = 'public_pb2'
  # @@protoc_insertion_point(class_scope:internal.Bitmap)
  ))
_sym_db.RegisterMessage(Bitmap)

Pair = _reflection.GeneratedProtocolMessageType('Pair', (_message.Message,), dict(
  DESCRIPTOR = _PAIR,
  __module__ = 'public_pb2'
  # @@protoc_insertion_point(class_scope:internal.Pair)
  ))
_sym_db.RegisterMessage(Pair)

SumCount = _reflection.GeneratedProtocolMessageType('SumCount', (_message.Message,), dict(
  DESCRIPTOR = _SUMCOUNT,
  __module__ = 'public_pb2'
  # @@protoc_insertion_point(class_scope:internal.SumCount)
  ))
_sym_db.RegisterMessage(SumCount)

Bit = _reflection.GeneratedProtocolMessageType('Bit', (_message.Message,), dict(
  DESCRIPTOR = _BIT,
  __module__ = 'public_pb2'
  # @@protoc_insertion_point(class_scope:internal.Bit)
  ))
_sym_db.RegisterMessage(Bit)

ColumnAttrSet = _reflection.GeneratedProtocolMessageType('ColumnAttrSet', (_message.Message,), dict(
  DESCRIPTOR = _COLUMNATTRSET,
  __module__ = 'public_pb2'
  # @@protoc_insertion_point(class_scope:internal.ColumnAttrSet)
  ))
_sym_db.RegisterMessage(ColumnAttrSet)

Attr = _reflection.GeneratedProtocolMessageType('Attr', (_message.Message,), dict(
  DESCRIPTOR = _ATTR,
  __module__ = 'public_pb2'
  # @@protoc_insertion_point(class_scope:internal.Attr)
  ))
_sym_db.RegisterMessage(Attr)

AttrMap = _reflection.GeneratedProtocolMessageType('AttrMap', (_message.Message,), dict(
  DESCRIPTOR = _ATTRMAP,
  __module__ = 'public_pb2'
  # @@protoc_insertion_point(class_scope:internal.AttrMap)
  ))
_sym_db.RegisterMessage(AttrMap)

QueryRequest = _reflection.GeneratedProtocolMessageType('QueryRequest', (_message.Message,), dict(
  DESCRIPTOR = _QUERYREQUEST,
  __module__ = 'public_pb2'
  # @@protoc_insertion_point(class_scope:internal.QueryRequest)
  ))
_sym_db.RegisterMessage(QueryRequest)

QueryResponse = _reflection.GeneratedProtocolMessageType('QueryResponse', (_message.Message,), dict(
  DESCRIPTOR = _QUERYRESPONSE,
  __module__ = 'public_pb2'
  # @@protoc_insertion_point(class_scope:internal.QueryResponse)
  ))
_sym_db.RegisterMessage(QueryResponse)

QueryResult = _reflection.GeneratedProtocolMessageType('QueryResult', (_message.Message,), dict(
  DESCRIPTOR = _QUERYRESULT,
  __module__ = 'public_pb2'
  # @@protoc_insertion_point(class_scope:internal.QueryResult)
  ))
_sym_db.RegisterMessage(QueryResult)

ImportRequest = _reflection.GeneratedProtocolMessageType('ImportRequest', (_message.Message,), dict(
  DESCRIPTOR = _IMPORTREQUEST,
  __module__ = 'public_pb2'
  # @@protoc_insertion_point(class_scope:internal.ImportRequest)
  ))
_sym_db.RegisterMessage(ImportRequest)


# @@protoc_insertion_point(module_scope)



# Copyright 2017 Pilosa Corp.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived
# from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
# DAMAGE.
#

import json

from .exceptions import PilosaError, ValidationError
from .validator import validate_index_name, validate_frame_name, validate_label

__all__ = ("TimeQuantum", "CacheType", "Schema", "Index", "PQLQuery", "PQLBatchQuery", "IntField", "RangeField", "Frame")

_TIME_FORMAT = "%Y-%m-%dT%H:%M"


class TimeQuantum:
    """Valid time quantum values for frames having support for that.
    
    * See: `Data Model <https://www.pilosa.com/docs/data-model/>`_
    """

    NONE = None
    YEAR = None
    MONTH = None
    DAY = None
    HOUR = None
    YEAR_MONTH = None
    MONTH_DAY = None
    DAY_HOUR = None
    YEAR_MONTH_DAY = None
    MONTH_DAY_HOUR = None
    YEAR_MONTH_DAY_HOUR = None

    def __init__(self, value):
        self.value = value

    def __str__(self):
        return self.value

    def __eq__(self, other):
        if isinstance(other, TimeQuantum):
            return self.value == other.value
        return False

TimeQuantum.NONE = TimeQuantum("")
TimeQuantum.YEAR = TimeQuantum("Y")
TimeQuantum.MONTH = TimeQuantum("M")
TimeQuantum.DAY = TimeQuantum("D")
TimeQuantum.HOUR = TimeQuantum("H")
TimeQuantum.YEAR_MONTH = TimeQuantum("YM")
TimeQuantum.MONTH_DAY = TimeQuantum("MD")
TimeQuantum.DAY_HOUR = TimeQuantum("DH")
TimeQuantum.YEAR_MONTH_DAY = TimeQuantum("YMD")
TimeQuantum.MONTH_DAY_HOUR = TimeQuantum("MDH")
TimeQuantum.YEAR_MONTH_DAY_HOUR = TimeQuantum("YMDH")


class CacheType:

    DEFAULT = None
    LRU = None
    RANKED = None

    def __init__(self, value):
        self.value = value

    def __str__(self):
        return self.value

    def __eq__(self, other):
        if isinstance(other, CacheType):
            return self.value == other.value
        return False

CacheType.DEFAULT = CacheType("")
CacheType.LRU = CacheType("lru")
CacheType.RANKED = CacheType("ranked")


class Schema:
    """Schema is a container for index objects"""

    def __init__(self):
        self._indexes = {}

    def __eq__(self, other):
        if id(self) == id(other):
            return True
        if not isinstance(other, self.__class__):
            return False
        return self._indexes == other._indexes

    def __ne__(self, other):
        return not self.__eq__(other)

    def index(self, name, column_label="columnID", time_quantum=TimeQuantum.NONE):
        """Returns an index object with the given name and options.

        If the index didn't exist in the schema, it is added to the schema.

        :param str name: index name
        :param str column_label: a valid column label. This field is deprecated and will be removed in a future release.
        :param pilosa.TimeQuantum time_quantum: Sets the time quantum. This field is deprecated and will be removed in a future release.
        :return: Index object

        * See `Data Model <https://www.pilosa.com/docs/data-model/>`_
        * See `Query Language <https://www.pilosa.com/docs/query-language/>`_
        * ``column_label`` field is deprecated.
        * ``time_quantum`` field is deprecated.
        """
        index = self._indexes.get(name)
        if index is None:
            index = Index(name, column_label, time_quantum)
            self._indexes[name] = index
        return index

    def _diff(self, other):
        result = Schema()
        for index_name, index in self._indexes.items():
            if index_name not in other._indexes:
                # if the index doesn't exist in the other schema, simply copy it
                result._indexes[index_name] = index.copy()
            else:
                # the index exists in the other schema; check the frames
                result_index = index.copy(frames=False)
                for frame_name, frame in index._frames.items():
                    # if the frame doesn't exist in the other scheme, copy it
                    if frame_name not in result_index._frames:
                        result_index._frames[frame_name] = frame.copy()
                # check whether we modified result index
                if len(result_index._frames) > 0:
                    result._indexes[index_name] = result_index

        return result


class Index:
    """The purpose of the Index is to represent a data namespace.
    
    You cannot perform cross-index queries. Column-level attributes are global to the Index.
    
    :param str name: index name
    :param str column_label: a valid column label
    :param pilosa.TimeQuantum time_quantum: Sets the time quantum

    * See `Data Model <https://www.pilosa.com/docs/data-model/>`_
    * See `Query Language <https://www.pilosa.com/docs/query-language/>`_    
    """

    def __init__(self, name, column_label="columnID", time_quantum=TimeQuantum.NONE):
        validate_index_name(name)
        validate_label(column_label)
        self.name = name
        self.column_label = column_label
        self.time_quantum = time_quantum
        self._frames = {}

    def __eq__(self, other):
        if id(self) == id(other):
            return True
        if not isinstance(other, self.__class__):
            return False
        return self._meta_eq(other) and \
               self._frames == other._frames

    def __ne__(self, other):
        return not self.__eq__(other)

    def _meta_eq(self, other):
        return self.name == other.name and \
               self.column_label == other.column_label and \
               self.time_quantum == other.time_quantum

    def copy(self, frames=True):
        index = Index(self.name, column_label=self.column_label, time_quantum=self.time_quantum)
        if frames:
            index._frames = dict((name, frame.copy()) for name, frame in self._frames.items())
        return index

    def frame(self, name, row_label="rowID", time_quantum=TimeQuantum.NONE,
              inverse_enabled=False, cache_type=CacheType.DEFAULT, cache_size=0, fields=None):
        """Creates a frame object with the specified name and defaults.
        
        :param str name: frame name
        :param str row_label: a valid row label. This field is deprecated and will be removed in a future release.
        :param pilosa.TimeQuantum time_quantum: Sets the time quantum for the frame. If a Frame has a time quantum, then Views are generated for each of the defined time segments.
        :param bool inverse_enabled:
        :param pilosa.CacheType cache_type: ``CacheType.DEFAULT``, ``CacheType.LRU`` or ``CacheType.RANKED``
        :param int cache_size: Values greater than 0 sets the cache size. Otherwise uses the default cache size
        :param list(IntField) fields: List of ``IntField`` objects. E.g.: ``[IntField.int("rate", 0, 100)]``
        :return: Pilosa frame
        :rtype: pilosa.Frame

        * ``row_label`` field is deprecated.
        """
        frame = self._frames.get(name)
        if frame is None:
            frame = Frame(self, name, row_label, time_quantum,
                          inverse_enabled, cache_type, cache_size, fields or [])
            self._frames[name] = frame
        return frame

    def raw_query(self, query):
        """Creates a raw query.
        
        Note that the query is not validated before sending to the server.
        
        :param str query:
        :return: Pilosa query
        :rtype: pilosa.PQLQuery
        """
        return PQLQuery(query, self)

    def batch_query(self, *queries):
        """Creates a batch query.
        
        :param pilosa.PQLQuery queries: the queries in the batch
        :return: Pilosa batch query
        :rtype: pilosa.PQLBatchQuery
        """
        q = PQLBatchQuery(self)
        q.add(*queries)
        return q

    def union(self, *bitmaps):
        """Creates a ``Union`` query.
        
        ``Union`` performs a logical OR on the results of each BITMAP_CALL query passed to it.
        
        :param pilosa.PQLBitmapQuery bitmaps: 0 or more bitmap queries to union
        :return: Pilosa bitmap query
        :rtype: pilosa.PQLBitmapQuery
        """
        return self._bitmap_op("Union", bitmaps)

    def intersect(self, *bitmaps):
        """Creates an ``Intersect`` query.

        ``Intersect`` performs a logical AND on the results of each BITMAP_CALL query passed to it.
        
        :param pilosa.PQLBitmapQuery bitmaps: 1 or more bitmap queries to intersect
        :return: Pilosa bitmap query
        :rtype: pilosa.PQLBitmapQuery
        :raise PilosaError: if the number of bitmaps is less than 1
        """
        if len(bitmaps) < 1:
            raise PilosaError("Number of bitmap queries should be greater than or equal to 1")
        return self._bitmap_op("Intersect", bitmaps)

    def difference(self, *bitmaps):
        """Creates a ``Difference`` query.

        ``Difference`` returns all of the bits from the first BITMAP_CALL argument passed to it,
        without the bits from each subsequent BITMAP_CALL.
        
        :param pilosa.PQLBitmapQuery bitmaps: 1 or more bitmap queries to differentiate
        :return: Pilosa bitmap query
        :rtype: pilosa.PQLBitmapQuery
        :raise PilosaError: if the number of bitmaps is less than 1
        """
        if len(bitmaps) < 1:
            raise PilosaError("Number of bitmap queries should be greater than or equal to 1")
        return self._bitmap_op("Difference", bitmaps)

    def xor(self, *bitmaps):
        """Creates a ``Xor`` query.

        :param pilosa.PQLBitmapQuery bitmaps: 2 or more bitmap queries to xor
        :return: Pilosa bitmap query
        :rtype: pilosa.PQLBitmapQuery
        :raise PilosaError: if the number of bitmaps is less than 2
        """
        if len(bitmaps) < 2:
            raise PilosaError("Number of bitmap queries should be greater than or equal to 2")
        return self._bitmap_op("Xor", bitmaps)

    def count(self, bitmap):
        """Creates a Count query.
        
        ``Count`` returns the number of set bits in the BITMAP_CALL passed in.
        
        :param pilosa.PQLQuery bitmap: the bitmap query
        :return: Pilosa query
        :rtype: pilosa.PQLQuery
        """
        return PQLQuery(u"Count(%s)" % bitmap.serialize(), self)

    def set_column_attrs(self, column_id, attrs):
        """Creates a SetColumnAttrs query.
        
        ``SetColumnAttrs`` associates arbitrary key/value pairs with a column in an index.
        
        Following object types are accepted:
        
        * int
        * str
        * bool
        * float
        
        :param int column_id:
        :param dict attrs: column attributes
        :return: Pilosa query
        :rtype: pilosa.PQLQuery        
        """
        attrs_str = _create_attributes_str(attrs)
        return PQLQuery(u"SetColumnAttrs(%s=%d, %s)" %
                        (self.column_label, column_id, attrs_str), self)

    def _bitmap_op(self, name, bitmaps):
        return PQLQuery(u"%s(%s)" % (name, u", ".join(b.serialize() for b in bitmaps)), self)


class Frame:
    """Frames are used to segment and define different functional characteristics within your entire index.
    
    You can think of a Frame as a table-like data partition within your Index.
    Row-level attributes are namespaced at the Frame level.
    
    Do not create a Frame object directly. Instead, use ``pilosa.Index.frame`` method.
        
    * See `Data Model <https://www.pilosa.com/docs/data-model/>`_
    * See `Query Language <https://www.pilosa.com/docs/query-language/>`_    
    """

    def __init__(self, index, name, row_label, time_quantum, inverse_enabled,
                 cache_type, cache_size, fields):
        validate_frame_name(name)
        validate_label(row_label)
        self.index = index
        self.name = name
        self.time_quantum = time_quantum
        self.inverse_enabled = inverse_enabled
        self.cache_type = cache_type
        self.cache_size = cache_size
        self.row_label = row_label
        self.column_label = index.column_label
        self.fields = fields
        self.range_fields = {}

    def __eq__(self, other):
        if id(self) == id(other):
            return True
        if not isinstance(other, self.__class__):
            return False

        # Note that we skip comparing the frames of the indexes by using index._meta_eq
        # in order to avoid a call cycle
        return self.name == other.name and \
               self.index._meta_eq(other.index) and \
               self.row_label == other.row_label and \
               self.time_quantum == other.time_quantum and \
               self.inverse_enabled == other.inverse_enabled and \
               self.cache_type == other.cache_type and \
               self.cache_size == other.cache_size and \
               self.fields == other.fields

    def __ne__(self, other):
        return not self.__eq__(other)

    def copy(self):
        return Frame(self.index, self.name, self.row_label, self.time_quantum,
                     self.inverse_enabled, self.cache_type, self.cache_size, self.fields)

    def bitmap(self, row_id):
        """Creates a Bitmap query.
        
        Bitmap retrieves the indices of all the set bits in a row or column based on whether the row label or column label is given in the query. It also retrieves any attributes set on that row or column.
        
        This variant of Bitmap query uses the row label.
        
        :param int row_id:
        :return: Pilosa bitmap query
        :rtype: pilosa.PQLBitmapQuery
        """
        return PQLQuery(u"Bitmap(%s=%d, frame='%s')" % (self.row_label, row_id, self.name),
                        self.index)

    def inverse_bitmap(self, column_id):
        """Creates a Bitmap query.

        ``Bitmap`` retrieves the indices of all the set bits in a row or column based on whether the row label or column label is given in the query. It also retrieves any attributes set on that row or column.

        This variant of Bitmap query uses the column label.

        :param int column_id:
        :return: Pilosa bitmap query
        :rtype: pilosa.PQLBitmapQuery
        """
        return PQLQuery(u"Bitmap(%s=%d, frame='%s')" % (self.column_label, column_id, self.name),
                        self.index)

    def setbit(self, row_id, column_id, timestamp=None):
        """Creates a SetBit query.
        
        ``SetBit`` assigns a value of 1 to a bit in the binary matrix, thus associating the given row in the given frame with the given column.
        
        :param int row_id:
        :param int column_id:
        :param pilosa.TimeStamp timestamp:
        :return: Pilosa query
        :rtype: pilosa.PQLQuery
        """
        ts = ", timestamp='%s'" % timestamp.strftime(_TIME_FORMAT) if timestamp else ''
        return PQLQuery(u"SetBit(%s=%d, frame='%s', %s=%d%s)" % \
                        (self.row_label, row_id, self.name, self.column_label, column_id, ts),
                        self.index)

    def clearbit(self, row_id, column_id):
        """Creates a ClearBit query.
        
        ``ClearBit`` assigns a value of 0 to a bit in the binary matrix, thus disassociating the given row in the given frame from the given column.
        
        :param int row_id:
        :param int column_id:
        :return: Pilosa query
        :rtype: pilosa.PQLQuery
        """
        return PQLQuery(u"ClearBit(%s=%d, frame='%s', %s=%d)" % \
                        (self.row_label, row_id, self.name, self.column_label, column_id),
                        self.index)

    def topn(self, n, bitmap=None, field="", *values):
        """Creates a TopN query.

        ``TopN`` returns the id and count of the top n bitmaps (by count of bits) in the frame.

        * see: `TopN Query <https://www.pilosa.com/docs/query-language/#topn>`_

        :param int n: number of items to return
        :param pilosa.PQLBitmapQuery bitmap: a PQL Bitmap query
        :param field str field: field name
        :param object values: filter values to be matched against the field
        """
        return self._topn(n, bitmap, field, False, *values)

    def inverse_topn(self, n, bitmap=None, field="", *values):
        """Creates a TopN query.

        ``TopN`` returns the id and count of the top n bitmaps (by count of bits) in the frame.

        This version sets `inverse=true`.

        * see: `TopN Query <https://www.pilosa.com/docs/query-language/#topn>`_

        :param int n: number of items to return
        :param pilosa.PQLBitmapQuery bitmap: a PQL Bitmap query
        :param field str field: field name
        :param object values: filter values to be matched against the field
        """
        return self._topn(n, bitmap, field, True, *values)

    def _topn(self, n, bitmap=None, field="", inverse=False, *values):
        parts = ["frame='%s'" % self.name, "n=%d" % n, "inverse=%s" % ('true' if inverse else 'false')]
        if bitmap:
            parts.insert(0, bitmap.serialize())
        if field:
            validate_label(field)
            values_str = json.dumps(values, separators=(',', ': '))
            parts.extend(["field='%s'" % field, "filters=%s" % values_str])
        qry = u"TopN(%s)" % ", ".join(parts)
        return PQLQuery(qry, self.index)

    def range(self, row_id, start, end):
        """Creates a Range query.

        Similar to ``Bitmap``, but only returns bits which were set with timestamps between the given start and end timestamps.

        * see: `Range Query <https://www.pilosa.com/docs/query-language/#range>`_

        :param int row_id:
        :param datetime.datetime start: start timestamp
        :param datetime.datetime end: end timestamp
        """
        return self._range(self.row_label, row_id, start, end)

    def inverse_range(self, column_id, start, end):
        """Creates a Range query.

        Similar to ``Bitmap``, but only returns bits which were set with timestamps between the given start and end timestamps.


        :param int column_id:
        :param datetime.datetime start: start timestamp
        :param datetime.datetime end: end timestamp
        """
        return self._range(self.column_label, column_id, start, end)

    def _range(self, label, rowcol_id, start, end):
        start_str = start.strftime(_TIME_FORMAT)
        end_str = end.strftime(_TIME_FORMAT)
        return PQLQuery(u"Range(%s=%d, frame='%s', start='%s', end='%s')" %
                        (label, rowcol_id, self.name, start_str, end_str),
                        self.index)

    def set_row_attrs(self, row_id, attrs):
        """Creates a SetRowAttrs query.
        
        ``SetRowAttrs`` associates arbitrary key/value pairs with a row in a frame.
        
        Following object types are accepted:
        
        * int
        * str
        * bool
        * float
        
        :param int row_id:
        :param dict attrs: row attributes
        :return: Pilosa query
        :rtype: pilosa.PQLQuery        
        """
        attrs_str = _create_attributes_str(attrs)
        return PQLQuery(u"SetRowAttrs(%s=%d, frame='%s', %s)" %
                        (self.row_label, row_id, self.name, attrs_str),
                        self.index)

    def field(self, name):
        """Returns a _RangeField object with the given name.

        :param name: field name
        :return: _RangeField object
        :rtype: RangeField
        """
        field = self.range_fields.get(name)
        if not field:
            validate_label(name)
            field = RangeField(self, name)
            self.range_fields[name] = field
        return field

    def _get_options_string(self):
        data = {"rowLabel": self.row_label}
        if self.inverse_enabled:
            data["inverseEnabled"] = True
        if self.time_quantum != TimeQuantum.NONE:
            data["timeQuantum"] = str(self.time_quantum)
        if self.cache_type != CacheType.DEFAULT:
            data["cacheType"] = str(self.cache_type)
        if self.cache_size > 0:
            data["cacheSize"] = self.cache_size
        if self.fields:
            data["rangeEnabled"] = True
            data["fields"] = [f.attrs for f in self.fields]
        return json.dumps({"options": data}, sort_keys=True)


class PQLQuery:

    def __init__(self, pql, index):
        self.pql = pql
        self.index = index

    def serialize(self):
        return self.pql


def _create_attributes_str(attrs):
    kvs = []
    try:
        for k, v in attrs.items():
            # TODO: make key use its own validator
            validate_label(k)
            kvs.append("%s=%s" % (k, json.dumps(v)))
        return ", ".join(sorted(kvs))
    except TypeError:
        raise PilosaError("Error while converting values")


class PQLBatchQuery:

    def __init__(self, index):
        self.index = index
        self.queries = []

    def add(self, *queries):
        self.queries.extend(queries)

    def serialize(self):
        return u''.join(q.serialize() for q in self.queries)


class IntField:

    def __init__(self, attrs):
        self.attrs = attrs

    @classmethod
    def int(cls, name, min=0, max=100):
        validate_label(name)
        if max <= min:
            raise ValidationError("Max should be greater than min for int fields")
        return cls({
            "name": name,
            "type": "int",
            "min": min,
            "max": max
        })


class RangeField:

    def __init__(self, frame, name):
        self.frame_name = frame.name
        self.name = name
        self.index = frame.index

    def lt(self, n):
        """Creates a Range query with less than (<) condition.

        :param n: The value to compare
        :return: a PQL query
        :rtype: PQLQuery
        """
        return self._binary_operation("<", n)

    def lte(self, n):
        """Creates a Range query with less than or equal (<=) condition.

        :param n: The value to compare
        :return: a PQL query
        :rtype: PQLQuery
        """
        return self._binary_operation("<=", n)

    def gt(self, n):
        """Creates a Range query with greater than (>) condition.

        :param n: The value to compare
        :return: a PQL query
        :rtype: PQLQuery
        """
        return self._binary_operation(">", n)

    def gte(self, n):
        """Creates a Range query with greater than or equal (>=) condition.

        :param n: The value to compare
        :return: a PQL query
        :rtype: PQLQuery
        """
        return self._binary_operation(">=", n)

    def equals(self, n):
        """Creates a Range query with equals (==) condition.

        :param n: The value to compare
        :return: a PQL query
        :rtype: PQLQuery
        """
        return self._binary_operation("==", n)

    def not_equals(self, n):
        """Creates a Range query with not equals (!=) condition.

        :param n: The value to compare
        :return: a PQL query
        :rtype: PQLQuery
        """
        return self._binary_operation("!=", n)

    def not_null(self):
        """Creates a Range query with not null (!= null) condition.

        :return: a PQL query
        :rtype: PQLQuery
        """
        q = u"Range(frame='%s', %s != null)" % (self.frame_name, self.name)
        return PQLQuery(q, self.index)

    def between(self, a, b):
        """Creates a Range query with between (><) condition.

        :param a: Closed range start
        :param b: Closed range end
        :return: a PQL query
        :rtype: PQLQuery
        """
        q = u"Range(frame='%s', %s >< [%d,%d])" % (self.frame_name, self.name, a, b)
        return PQLQuery(q, self.index)

    def sum(self, bitmap):
        """Creates a Sum query.

        :param bitmap: The bitmap query to use.
        :return: a PQL query
        :rtype: PQLQuery
        """
        q = u"Sum(%s, frame='%s', field='%s')" % (bitmap.serialize(), self.frame_name, self.name)
        return PQLQuery(q, self.index)

    def set_value(self, column_id, value):
        """Creates a SetFieldValue query.

        :param column_id: column ID
        :param value: the value to assign to the field
        :return: a PQL query
        :rtype: PQLQuery
        """
        q = u"SetFieldValue(frame='%s', %s=%d, %s=%d)" % \
            (self.frame_name, self.index.column_label, column_id, self.name, value)
        return PQLQuery(q, self.index)

    def _binary_operation(self, op, n):
        q = u"Range(frame='%s', %s %s %d)" % (self.frame_name, self.name, op, n)
        return PQLQuery(q, self.index)



# Copyright 2017 Pilosa Corp.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived
# from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
# DAMAGE.
#

import re

from pilosa.exceptions import ValidationError

__all__ = ("valid_index_name", "validate_index_name", "valid_frame_name",
           "validate_frame_name", "valid_label", "validate_label")


__INDEX_NAME = re.compile(r"^[a-z][a-z0-9_-]*$")
__FRAME_NAME = re.compile(r"^[a-z][a-z0-9_-]*$")
__LABEL = re.compile(r"^[a-zA-Z][a-zA-Z0-9_-]*$")
__MAX_INDEX_NAME = 64
__MAX_FRAME_NAME = 64
__MAX_LABEL = 64


def valid_index_name(index_name):
    if len(index_name) > __MAX_INDEX_NAME:
        return False
    return bool(__INDEX_NAME.match(index_name))


def validate_index_name(index_name):
    if not valid_index_name(index_name):
        raise ValidationError("Invalid index name: %s" % index_name)


def valid_frame_name(frame_name):
    if len(frame_name) > __MAX_FRAME_NAME:
        return False
    return bool(__FRAME_NAME.match(frame_name))


def validate_frame_name(frame_name):
    if not valid_frame_name(frame_name):
        raise ValidationError("Invalid frame name: %s" % frame_name)


def valid_label(label):
    if len(label) > __MAX_LABEL:
        return False
    return bool(__LABEL.match(label))


def validate_label(label):
    if not valid_label(label):
        raise ValidationError("Invalid label: %s" % label)



# Copyright 2017 Pilosa Corp.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived
# from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
# DAMAGE.
#

import itertools
from collections import namedtuple

from pilosa.exceptions import PilosaError

__all__ = ("Bit", "csv_bit_reader")

Bit = namedtuple("Bit", "row_id column_id timestamp")


def csv_bit_reader(file_obj, timefunc=int):
    """
    Reads bits from the given file-like object.

    Each line of the file-like object should correspond to a single bit and must be in the following form:
    rowID,columnID[,timestamp]

    :param file_obj:
    :param timefunc: optional time parsing function, defaults to int
    :return: a generator
    """
    for line in file_obj:
        line = line.strip()
        if not line:
            continue
        parts = line.split(",")
        if len(parts) == 2:
            try:
                bit = Bit(row_id=int(parts[0]), column_id=int(parts[1]), timestamp=0)
            except ValueError:
                raise PilosaError("Invalid CSV line: %s", line)
        elif len(parts) == 3:
            try:
                bit = Bit(row_id=int(parts[0]), column_id=int(parts[1]), timestamp=timefunc(parts[2]))
            except ValueError:
                raise PilosaError("Invalid CSV line: %s", line)
        else:
            raise PilosaError("Invalid CSV line: %s", line)
        yield bit


def batch_bits(reader, batch_size):
    slice_width = 1048576
    while 1:
        batch = list(itertools.islice(reader, batch_size))
        if not batch:
            break
        bit_groups = {}
        for bit in batch:
            bit_groups.setdefault(bit.column_id // slice_width, []).append(bit)
        for slice_bit_group in bit_groups.items():
            yield slice_bit_group




# Copyright 2017 Pilosa Corp.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived
# from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
# DAMAGE.
#

from codecs import open
from os import path

from setuptools import setup

here = path.abspath(path.dirname(__file__))

try:
    import pypandoc
    long_description = pypandoc.convert_file("README.md", "rst")
except ImportError:
    # long_description is required only during the release to PyPI
    long_description = u''

exec(open(path.join(here, 'pilosa/version.py'), 'r').read())

setup(
    name='pilosa',
    version=_get_version_setup(),
    description='Python client library for Pilosa',
    long_description=long_description,
    url='https://github.com/pilosa/python-pilosa',
    author='Pilosa Engineering',
    author_email='dev@pilosa.com',
    license='BSD',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Operating System :: OS Independent',
        'License :: OSI Approved :: BSD License',
    ],

    keywords='pilosa,pql',
    packages=['pilosa', 'pilosa.internal'],

    setup_requires=['pytest-runner'],
    install_requires=['urllib3', 'protobuf'],
    tests_require=['pytest', 'mock', 'coverage', 'pytest-cov'],
)



# Copyright 2017 Pilosa Corp.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived
# from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
# DAMAGE.
#

__all__ = ("PilosaError", "ValidationError", "PilosaURIError", "IndexExistsError", "FrameExistsError")


class PilosaError(Exception):
    pass


class ValidationError(PilosaError):
    pass


class PilosaURIError(PilosaError):
    pass


class IndexExistsError(PilosaError):
    pass


class FrameExistsError(PilosaError):
    pass

